/****************************************************************************
**
** Copyright (C) 2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This file is part of the Qt Components project.
**
** $QT_BEGIN_LICENSE:BSD$
** You may use this file under the terms of the BSD license as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor
**     the names of its contributors may be used to endorse or promote
**     products derived from this software without specific prior written
**     permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
** $QT_END_LICENSE$
**
****************************************************************************/

#include "sdeclarativeframe.h"
#include "sframepool.h"
#include "../sdeclarative.h"

#include <qcoreapplication.h>
#include <qpainter.h>
#include <qgraphicsscene.h>

//#define Q_DEBUG_FRAME
#ifdef Q_DEBUG_FRAME
#include <qdebug.h>
#endif

class SDeclarativeFramePrivate
{
public:
    SDeclarativeFramePrivate() :
        frameType(SDeclarativeFrame::Undefined),
        frameLoadAttempted(false) {}

    ~SDeclarativeFramePrivate() {}

    void releaseFromFramePool();

public:
    QPixmap framePixmap;
    QString frameName;
    QString fileName;
    SDeclarativeFrame::FrameType frameType;
    bool frameLoadAttempted;
    QSize loadedSize;
};

void SDeclarativeFramePrivate::releaseFromFramePool()
{
    // Release previously loaded pixmap from frame pool
    SFramePool::release(fileName, frameType, loadedSize);
    framePixmap = QPixmap();
    frameLoadAttempted = false;
}

SDeclarativeFrame::SDeclarativeFrame(QDeclarativeItem *parent) :
    QDeclarativeItem(parent),
    d_ptr(new SDeclarativeFramePrivate)
{
    setFlag(QGraphicsItem::ItemHasNoContents, false);
}

SDeclarativeFrame::~SDeclarativeFrame()
{
    Q_D(SDeclarativeFrame);
    // Release from icon pool
    if (d->frameLoadAttempted)
        d->releaseFromFramePool();
}

QString SDeclarativeFrame::frameName() const
{
    Q_D(const SDeclarativeFrame);
    return d->frameName;
}

void SDeclarativeFrame::setFrameName(const QString &frameName)
{
    Q_D(SDeclarativeFrame);
    if (frameName != d->frameName) {
        if (d->frameLoadAttempted) {
            // Release previous one from frame pool
            d->releaseFromFramePool();
        }

        d->fileName = SDeclarative::resolveIconFileName(frameName);
        d->frameName = frameName;
        update();
        emit frameNameChanged(frameName);
    }
}

SDeclarativeFrame::FrameType SDeclarativeFrame::frameType() const
{
    Q_D(const SDeclarativeFrame);
    return d->frameType;
}

void SDeclarativeFrame::setFrameType(SDeclarativeFrame::FrameType type)
{
    Q_D(SDeclarativeFrame);
    if (d->frameType != type) {
        if (d->frameLoadAttempted) {
            // Release previous one from frame pool
            d->releaseFromFramePool();
        }

        d->frameType = type;
        emit frameTypeChanged(type);
    }
}

void SDeclarativeFrame::paint(QPainter *painter, const QStyleOptionGraphicsItem *, QWidget *)
{
    Q_D(SDeclarativeFrame);
    QSize rectSize(width(), height());

    if (rectSize.isEmpty())
        return;

    // If the rect size has changed, frame needs to be reloaded
    if (d->frameLoadAttempted && d->loadedSize != rectSize)
        d->releaseFromFramePool();

    // Load pixmap now that rect size is known
    if (!d->frameLoadAttempted && !d->fileName.isEmpty()) {
        // Get frame pixmap from application's frame pool
        d->framePixmap = SFramePool::get(d->fileName, d->frameType, rectSize);
        d->frameLoadAttempted = true;
        d->loadedSize = rectSize;
    }

    if (!d->framePixmap.isNull())
        painter->drawPixmap(QPoint(0, 0), d->framePixmap);

#ifdef Q_DEBUG_FRAME
    qDebug() << "SDeclarativeFrame::paint";
    qDebug() << "framename = " << frameName();

    qDebug() << "width = " << width();
    qDebug() << "height = " << height();
#endif
}
