/****************************************************************************
**
** Copyright (C) 2011 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This file is part of the Qt Components project.
**
** $QT_BEGIN_LICENSE:BSD$
** You may use this file under the terms of the BSD license as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor
**     the names of its contributors may be used to endorse or promote
**     products derived from this software without specific prior written
**     permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
** $QT_END_LICENSE$
**
****************************************************************************/

#include "siconpool.h"

#include <QPainter>
#include <QSvgRenderer>
#include <QPixmap>
#include <QSizeF>
#include <QHash>

//#define Q_DEBUG_ICON
#ifdef Q_DEBUG_ICON
#include <QDebug>
#endif

/*!
  \class SIconPoolKey
 */
struct SIconPoolKey {
public:
    SIconPoolKey(const QString &filename, const QSize &size, Qt::AspectRatioMode mode)
        : mFilename(filename), mSize(size), mMode(mode) {}

    bool operator==(const SIconPoolKey &other) const {
        return other.mFilename == mFilename && other.mSize == mSize && other.mMode == mMode;
    }

    QString mFilename;
    QSize mSize;
    Qt::AspectRatioMode mMode;
};

/*!
  \class SIconPoolValue
 */
struct SIconPoolValue {
    SIconPoolValue()
        : mPixmap(), mRefCount(0) {}
    SIconPoolValue(const QPixmap &pixmap)
        : mPixmap(pixmap), mRefCount(1) {}
    QPixmap mPixmap;
    int mRefCount;
};

uint qHash(const SIconPoolKey &key)
{
    return qHash(key.mFilename);
}

typedef QHash<SIconPoolKey, SIconPoolValue> SIconPoolData;
Q_GLOBAL_STATIC(SIconPoolData, poolData);

QPixmap SIconPool::get(const QString &filename, const QSize &size, Qt::AspectRatioMode mode)
{
    QPixmap pixmap;

    if (!filename.isEmpty() && !size.isEmpty()) {
        SIconPoolKey key(filename, size, mode);
        SIconPoolData *pool = poolData();

        if (pool->contains(key)) {
            SIconPoolValue value = pool->value(key);
            ++value.mRefCount;
            pixmap = value.mPixmap;
            pool->insert(key, value);
        } else {
            pixmap = loadIcon(filename, size, mode);
            pool->insert(key, SIconPoolValue(pixmap));
        }
#ifdef Q_DEBUG_ICON
        qDebug() << "SIconPool::get()" << key.mFilename << pool->value(key).mRefCount;
#endif
    }
    return pixmap;
}

/**
* This is just a temp solution to get something visible - final icon loading architecture is worked on in SVG team.
*/
QPixmap SIconPool::loadIcon(const QString &filename, const QSize &size, Qt::AspectRatioMode mode)
{
    QPixmap pm;
    // SVG? Use QSvgRenderer
    if (filename.endsWith(".svg")) {
        QSvgRenderer *renderer = getSvgRenderer(filename);

        if (renderer->isValid()) {
            QSize renderSize = renderer->defaultSize();

            // a bit of a hack: should we expose the aspect ratio flag?
            if (filename.contains("screen_bg")) {
                mode = Qt::KeepAspectRatioByExpanding;
            }
            renderSize.scale(size, mode);

            pm = QPixmap(renderSize);
            pm.fill(QColor(0, 0, 0, 0));
            QPainter painter(&pm);
            renderer->render(&painter, QRectF(QPointF(), renderSize));
        }
    } else {
        // Otherwise load with QPixmap
        pm.load(filename);
        if (!pm.isNull()) {
            pm = pm.scaled(size, mode, Qt::SmoothTransformation);
        }
    }

#ifdef Q_DEBUG_ICON
    if (pm.isNull()) {
        qDebug() << "Fail to load icon: " << filename;
    }
#endif

    return pm;
}

void SIconPool::release(const QString &filename, const QSize &size, Qt::AspectRatioMode mode)
{
    SIconPoolKey key(filename, size, mode);
    SIconPoolData *pool = poolData();
    if (pool->contains(key)) {
        SIconPoolValue value = pool->value(key);
        // Delete if last reference
        if (!--value.mRefCount) {
            pool->remove(key);
        } else {
            // Update dereferenced value in pool
            pool->insert(key, value);
        }

#ifdef Q_DEBUG_ICON
        qDebug() << "SIconPool::release()" << key.mFilename << value.mRefCount;
#endif
    }
}

QSize SIconPool::defaultSize(const QString &filename)
{
    QSize defSize;

    // Get the default size from svg renderer or pixmap size
    if (!filename.isEmpty()) {
        // SVG? Use QSvgRenderer
        if (filename.endsWith(".svg")) {
            QSvgRenderer *svgRenderer = getSvgRenderer(filename);
            if (svgRenderer->isValid()) {
                defSize = svgRenderer->defaultSize();
            }
        } else {
            // Otherwise load with QPixmap
            QPixmap pixmap;
            pixmap.load(filename);
            defSize = pixmap.size();
        }
    }

    return defSize;
}

QSvgRenderer *SIconPool::getSvgRenderer(const QString &filename)
{
    static QString lastSvgFileName;
    static QSvgRenderer *lastSvgRenderer = 0;

    if (lastSvgFileName == filename)
        return lastSvgRenderer;

    delete lastSvgRenderer;
    lastSvgRenderer = new QSvgRenderer(filename);
    lastSvgFileName = filename;
    return lastSvgRenderer;
}

#ifdef ICON_POOL_UNIT_TEST

int SIconPool::totalCount()
{
    SIconPoolData *pool = poolData();
    return pool->count();
}

int SIconPool::count(const QString &filename, const QSize &size, Qt::AspectRatioMode mode)
{
    SIconPoolKey key(filename, size, mode);
    SIconPoolData *pool = poolData();
    if (pool->contains(key)) {
        SIconPoolValue value = pool->value(key);
        return value.mRefCount;
    } else {
        return 0;
    }
}

#endif // ICON_POOL_UNIT_TEST
